"""
/***************************************************************************
SimpleSvg
A QGIS plugin
Create simple SVG from current view, editable with InkScape
                             -------------------
begin                : 2011-06-16
copyright            : (C) 2011 by Richard Duivenvoorde
email                : richard@duif.net 
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
# Import the PyQt and QGIS libraries
from PyQt4.QtCore import * 
from PyQt4.QtGui import *
from qgis.core import *
# Initialize Qt resources from file resources.py
import resources
# Import the code for the dialog
from SimpleSvgDialog import SimpleSvgDialog

SVG_TYPE_PATH = 1
SVG_TYPE_SHAPE = 2

# next:
# using short notation ?
# using styles instead of inline stuff ? Mmm, apparently no css-style implementation in inkscape yet ...??
# trying to calculate de viewport
# see if it's possible to use world coordinates instead of pixels
# make 'skipping abundant points' an option
# adding raster layers als inline images


# http://www.carto.net/papers/svg/samples/
# uit: http://www.carto.net/papers/svg/samples/viewbox.shtml
# image includen:
#<g id="relief">
#   <image transform="matrix(0.876704, 0, 0, 0.8767, -2357.76, 1302.27)" x="0" y="-1472" width="5258" height="2642" xlink:href="x_austria_relief.png"/>
#</g>

# style definities
#<defs>
#  <style type="text/css">
#   <![CDATA[
#    .str7 {stroke:#1F1A17;stroke-width:3}
#    .str4 {stroke:#0093DD;stroke-width:3}
#    .str5 {stroke:#0093DD;stroke-width:4}
#    .str3 {stroke:#0093DD;stroke-width:5}
#    .str2 {stroke:#0093DD;stroke-width:6}
#    .str1 {stroke:#0093DD;stroke-width:8}
#    .str6 {stroke:#ee9090;stroke-width:14}
#    .str0 {stroke:#DA251D;stroke-width:21}
#    .fil0 {fill:none}
#    .fil2 {fill:#1F1A17}
#    .fil1 {fill:#C4E5FA}
#    .fil3 {fill:#FFFFFF}
#    .fnt1 {font-weight:normal;font-size:125px;font-family:'Arial';text-rendering:optimizeLegibility;}
#    .fnt0 {font-weight:normal;font-size:139px;font-family:'Arial';text-decoration:underline;text-rendering:optimizeLegibility;}
#   ]]>
#  </style>
# </defs>
  
# short notation
# <path class="fil0 str6" d="M113 1084c2,9 14,23 4,29 -10,5 -28,-14 -39,-9 -11,5 -8,21 -12,31"/>

# (point) symbols
# http://www.carto.net/papers/svg/samples/symbol.shtml

class SimpleSvg:

  def __init__(self, iface):
    # Save reference to the QGIS interface
    self.iface = iface
    self.filePath = "/home/richard/temp/svgtest.svg"
    self.svgType = SVG_TYPE_PATH
    self.strokeLineJoin = 'round' # miter, round, bevel

  def initGui(self):  
      # Create action that will start plugin configuration
    self.action = QAction(QIcon(":/plugins/simplesvg/icon.png"), \
            "Save as SVG", self.iface.mainWindow())
    # connect the action to the run method
    QObject.connect(self.action, SIGNAL("activated()"), self.run) 

    # Add toolbar button and menu item
    self.iface.addToolBarIcon(self.action)
    self.iface.addPluginToMenu("&Save as SVG", self.action)

  def unload(self):
    # Remove the plugin menu item and icon
    self.iface.removePluginMenu("&Save as SVG",self.action)
    self.iface.removeToolBarIcon(self.action)

  # run method that performs all the real work
  def run(self):
    # create and show the dialog 
    dlg = SimpleSvgDialog()
    # show the dialog
    dlg.show()
    result = dlg.exec_() 
    # See if OK was pressed
    if result == 1: 
      #pass
      self.svgFilename = dlg.getFilePath()
      self.writeSVG()

  def writeSVG(self):
    #print "ok starting"

    # determine extent for later use (only write geoms that are at least partially contained)
    self.currentExtent = self.iface.mapCanvas().extent()

    svg = [""]
    svg.append('<?xml version="1.0" standalone="no"?>\n')
    svg.append('<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN" "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">\n')
    #svg.append('<svg viewBox = "0 0 800 800" version = "1.1">\n<g id="layers">\n')
    svg.append('<svg xmlns="http://www.w3.org/2000/svg" xmlns:inkscape="http://www.inkscape.org/namespaces/inkscape" xmlns:xlink="http://www.w3.org/1999/xlink" viewBox = "0 0 800 800" version = "1.1">\n')
    svg.append('<!-- svg generated using QGIS www.qgis.org -->\n')
    

    # for all visible layers, from bottom to top (-1)
    for i in range (self.iface.mapCanvas().layerCount()-1, -1, -1):
        # only visible layers are counted and found anyway ...
        #print self.iface.mapCanvas().layer(i).name()
        # print provider
        #print self.iface.mapCanvas().layer(i).dataProvider()
        # only vector data providers can be used to write svg (?? is it? we can try to insert a raster image?)
        # ??? how can we determine whether a layer is vector or raster based?
        layer = self.iface.mapCanvas().layer(i)
        if layer.type()==0 and hasattr(layer, 'isUsingRendererV2') and layer.isUsingRendererV2():
            QMessageBox.information(self.iface.mainWindow(), "Warning", "New Symbology layer found for layer '"+layer.name()+"'\n\nThe plugin cannot handle layer(s) which use 'New Symbology' yet.\n\nThis layer will be ignored in export.\n\nPlease change symbology of these layer(s) to 'Old Symbology' if you want this layer in svg.")
        elif layer.type()==0: # vector
          svg.extend(self.writeLayer(layer, False))
        elif layer.type()==1: # raster
          svg.extend(self.writeRaster(layer))
        # layers like OpenLayers/OpenStreetmap/Google are plugin layer: write as raster for now
        elif layer.type()==2: # plugin layer 
          svg.extend(self.writeRaster(layer))

    # now layers with labels
    for i in range (self.iface.mapCanvas().layerCount()-1, -1, -1):
        # only visible layers are counted and found anyway ...
        #print self.iface.mapCanvas().layer(i).name()
        # print provider
        #print self.iface.mapCanvas().layer(i).dataProvider()
        # only vector data providers can be used to write svg (?? is it? we can try to insert a raster image?)
        # ??? how can we determine whether a layer is vector or raster based?
        layer = self.iface.mapCanvas().layer(i)
        if layer.type()==0 and layer.hasLabelsEnabled(): # only vectors have labels
          svg.extend(self.writeLayer(layer, True))

    # qgis extent, usable for clipping in Inkscape
    svg.extend(self.writeExtent())

    #svg.append("</g>\n</svg>")
    svg.append("</svg>")
    file = open(self.svgFilename, "w")
    file.writelines(svg)

  def writeExtent(self):
    self.extentAsPoly = QgsGeometry();
    self.extentAsPoly = QgsGeometry.fromRect(self.currentExtent);
    svg = '<!-- QGIS extent for clipping, eg in Inkscape -->\n<g id="qgisviewbox" inkscape:groupmode="layer" inkscape:label="qgisviewbox" stroke="rgb(255,0,0)" stroke-width="1" fill="none" >\n'
    for ring in self.extentAsPoly.asPolygon():
        svg += '<path d="M '
        first = True
        for point in ring:
            pixpoint =  self.w2p(point.x(), point.y(), self.iface.mapCanvas().mapUnitsPerPixel(), self.currentExtent.xMinimum(), self.currentExtent.yMaximum())
            if not first:
                svg += 'L '
            svg += (str(pixpoint[0]) + ',' + str(pixpoint[1]) + ' ')
            first = False
        svg += '" />\n'
    svg += '</g>'
    return svg


  def writeRaster(self, layer):
    # hide all layers except 'layer' and save as png image in current directory
    # TODO? maybe inline it in svg?

    # save visibility of layers
    visibleList=self.iface.mapCanvas().layers()
    legend = self.iface.legendInterface()
    # set all layers invisible EXCEPT layer
    for lyr in visibleList:
      if lyr != layer:
        legend.setLayerVisible(lyr, False)
    imgName = layer.name()+'.png'
    imgPath= self.filePath[:self.filePath.rfind('/')+1]
    # save image next to svg but put it in Image tag only the local filename
    self.iface.mapCanvas().saveAsImage(imgPath+imgName)
    # <image y="-7.7685061" x="27.115078" id="image3890" xlink:href="nl.png" />
    svg = ['<g id="'+layer.name()+'" inkscape:groupmode="layer" inkscape:label="'+layer.name()+'">\n'];
    #svg.append('<image y="0" x="0" xlink:href="'+imgPath+imgName+'" />')
    svg.append('<image y="0" x="0" xlink:href="'+imgName+'" />')
    svg.append('</g><!-- end of raster layer -->')
    # now set earlier visible layers back to visible
    for lyr in visibleList:
      legend.setLayerVisible(lyr, True)
    return svg

  def writeLayer(self, layer, labels=False):
    # keep the current extent as a Geometry to be able to do some 'contains' tests later
    self.extentAsPoly = QgsGeometry();
    self.extentAsPoly = QgsGeometry.fromRect(self.currentExtent);
    # select features within current extent,
    #   with  ALL attributes, WITHIN currentExtent, WITH geom, AND using Intersect instead of bbox
    provider = layer.dataProvider();
    provider.select(provider.attributeIndexes(), self.currentExtent, True, True)
    # get a list of all selected features ids
#    selectedFeaturesIds = self.iface.activeLayer().selectedFeaturesIds()
    # set max progress bar to number of features (not very accurate with a lot of huge multipolygons)
    #self.imageMapPluginGui.setProgressBarMax(self.iface.activeLayer().featureCount())
    # or run over all features in current selection, just to determine the number of... (should be simpler ...)
#    count = 0
#    while self.provider.nextFeature(feature):
#        count = count + 1
#    self.imageMapPluginGui.setProgressBarMax(count)
#    progressValue = 0
    # it seems that a postgres provider is on the end of file now
    # we do the select again to set the pointer/cursor to 0 again ?
#    self.provider.select(self.provider.attributeIndexes(), currentExtent, True, True)
#    self.provider.rewind()

    # TRYING TO GROUP BY SYMBOL
    renderer = layer.renderer()
    symbols = renderer.symbols()
    symbolFeatureMap = dict.fromkeys(symbols, [])
    if renderer.name() == "Continuous Color":
        symbols = []
        symbolFeatureMap = dict()
    id=self.sanitizeStr(layer.name())
    id=str(id).lower()
    if labels:
        id=id+'_labels'
    svg = ['<g id="'+id+'" inkscape:groupmode="layer" inkscape:label="'+id+'">\n'];
    # now iterate through each feature and group by feature
    f = QgsFeature();
    while provider.nextFeature(f):
      feature = QgsFeature(f)
      #progressValue = progressValue+1
      #self.imageMapPluginGui.setProgressBarValue(progressValue)
      #if False: #self.selectedFeaturesOnly and feature.id() not in selectedFeaturesIds:
        # print "skipping %s " % feature.id()
        #None
      symbol = self.symbolForFeature(renderer, feature)
      #print "feature: %s  symbol: %s rgb: %s %s %s" % (feature,symbol,symbol.color().red(),symbol.color().green(),symbol.color().blue())
      if renderer.name() == "Continuous Color":
        symbols.append(symbol)
        if symbol in symbolFeatureMap:
            symbolFeatureMap[symbol]=[feature]
        else:
            symbolFeatureMap.update({symbol:[feature]})
      else:
        if len(symbolFeatureMap[symbol])==0:
          symbolFeatureMap[symbol]=[feature]
        else:
          symbolFeatureMap[symbol].append(feature)
    # now iterate over symbols
    id=id+'_'
    i=0
    for symbol in symbols:
      # symbol color
      svg.append('<g>\n'); 
      sc = symbol.color()
      color = 'rgb(%s,%s,%s)' % (sc.red(), sc.green(), sc.blue())
      # fill color: only non line features have fill color lines have 'none'
      sfc = symbol.fillColor()
      fillColor = 'rgb(%s,%s,%s)' % (sfc.red(), sfc.green(), sfc.blue())
      geom = feature.geometry()
      if geom.wkbType() == QGis.WKBLineString or geom.wkbType() == QGis.WKBMultiLineString:
        fillColor = 'none'
      # pen: in QT pen can be 0
      pen = symbol.pen().width()
      if pen<1:
        pen=0.5
      for feature in symbolFeatureMap[symbol]:
        i=i+1
        if not labels:
          svg.extend(self.writeFeature(feature, id+str(i), color, fillColor, pen))
        if labels:
          #labelField = layer.label().labelField(0) # only first field for now.
          labeltxt = layer.label().fieldValue(0, feature)
          svg.extend(self.label2svg(feature, id+str(i), self.symbolForFeature(renderer, feature), labeltxt))
      svg.append('</g><!-- end of symbol g -->\n');
    svg.append('</g><!-- end of id/layer g -->\n');
    return svg

  def sanitizeStr(self, string):
    # TODO: find the right way to do this
    return string.replace(' ','_').replace('/','_').replace(',','_').replace('.','_')

  def writeFeature(self, feature, fid, color, fillColor, pen):
    svg = ['']
    # <g>-element set's style attributes
    svg.append('<g id="' + fid + '" stroke="' + color + '" fill="' + fillColor + '" stroke-linejoin="' + self.strokeLineJoin + '" stroke-width="' + str(pen) + '">\n')
    geom=feature.geometry()
    currentExtent=self.currentExtent
    if geom.wkbType() == QGis.WKBPoint: # 1 = WKBPoint
        buffer = self.iface.mapCanvas().mapUnitsPerPixel()*5 # (plusminus 20pixel areas)
        # either create a buffer around this point
#        polygon = geom.buffer(buffer,0).asPolygon()
#        svg.extend(self.polygon2svg(feature, polygon, currentExtent, color, fillColor, pen))
        # or create a circle
        svg.extend(self.point2svg(feature, currentExtent, color, fillColor, pen))
    if geom.wkbType() == QGis.WKBPolygon: # 3 = WKBTYPE.WKBPolygon:
        polygon = geom.asPolygon()  # returns a list
        svg.extend(self.polygon2svg(feature, polygon, currentExtent, color, fillColor, pen))
    if geom.wkbType() == QGis.WKBMultiPolygon: # 6 = WKBTYPE.WKBMultiPolygon:
        multipolygon = geom.asMultiPolygon() # returns a list
        for polygon in multipolygon:
          svg.extend(self.polygon2svg(feature, polygon, currentExtent, color, fillColor, pen))
    if geom.wkbType() == QGis.WKBLineString: # 6 = WKBTYPE.WKBLineString:
        line = geom.asPolyline()  # returns a list of points
        svg.extend(self.line2svg(feature, line, currentExtent, color, fillColor, pen))
    if geom.wkbType() == QGis.WKBMultiLineString: # 6 = WKBTYPE.WKBLineString:
        multiline = geom.asMultiPolyline()  # returns a list of points
        for line in multiline:
            svg.extend(self.line2svg(feature, line, currentExtent, color, fillColor, pen))
    svg.append('</g>\n');
    return svg

  def point2svg(self, feature, currentExtent, color, fillColor, pen):
    point = feature.geometry().asPoint()
    xy =  self.w2p(point.x(), point.y(), self.iface.mapCanvas().mapUnitsPerPixel(), self.currentExtent.xMinimum(), self.currentExtent.yMaximum())
    # TODO take current extent into account
    svg = '<circle cx="'+str(xy[0])+'" cy="'+str(xy[1])+'" r="5" stroke="'+color+'" stroke-width="'+str(pen)+'" fill="'+fillColor+'"/>'
    return svg

  def label2svg(self, feature, fid, symbol, labelTxt):
    # <g> <text x="262.08704" y="523.79077">abc</text> </g>
    point = feature.geometry().centroid().asPoint()
    xy =  self.w2p(point.x(), point.y(), self.iface.mapCanvas().mapUnitsPerPixel(), self.currentExtent.xMinimum(), self.currentExtent.yMaximum())
    #svg = '<g> <text x="'+str(xy[0])+'" y="'+str(xy[1])+'">'+labelTxt+'</text> </g>\n'
    svg = '<text id="'+fid+'" x="'+str(xy[0])+'" y="'+str(xy[1])+'" label="label">'+labelTxt+'</text>\n'
    return svg

  def symbolForFeature(self, renderer, feature):
    #print renderer.name()
    default = None
    # OLD symbolisation:
    #   Graduated Symbol: every symbol has BOTH upper and lower bound/value
    #   Unique Value: every symbol has BOTH upper and lower value
    #   Continues Color: HAS lower(==value) but NO upper value, BUT has just two symbols: MIN and MAX color, see http://doc.qgis.org/head/qgscontinuouscolorrenderer_8cpp-source.html
    #   Single Simbol: just one symbol, return it
    if renderer.name() == "Single Symbol":   # there is just one symbol: return it
      return renderer.symbols()[0]
    if renderer.name() == "Continuous Color":
      #pass  # already done, should not come here !!
      minSymbol = renderer.symbols()[0]
      minValue = (minSymbol.lowerValue()).toDouble()[0]  # we know Continuous Color only works with numeric attributes
      minRed = minSymbol.fillColor().red()
      minGreen = minSymbol.fillColor().green()
      minBlue = minSymbol.fillColor().blue()
      maxSymbol = renderer.symbols()[1]
      maxValue = (maxSymbol.lowerValue()).toDouble()[0]  # we know Continuous Color only works with numeric attributes
      maxRed = maxSymbol.fillColor().red()
      maxGreen = maxSymbol.fillColor().green()
      maxBlue = maxSymbol.fillColor().blue()
      # create new symbol and set RGB according to calculation in http://doc.qgis.org/head/qgscontinuouscolorrenderer_8cpp-source.html
      symbol = QgsSymbol(minSymbol) # copy
      value = (feature.attributeMap()[renderer.classificationAttributes()[0]]).toDouble()[0]  # we know Continuous Color only works with numeric attributes
      if (maxValue - minValue) <> 0:
        red = int ( maxRed * ( value - minValue ) / ( maxValue - minValue ) + minRed * ( maxValue - value ) / ( maxValue - minValue ) )
        green = int ( maxGreen * ( value - minValue ) / ( maxValue - minValue ) + minGreen * ( maxValue - value ) / ( maxValue - minValue ) )
        blue =  int ( maxBlue * ( value - minValue ) / ( maxValue - minValue ) + minBlue * ( maxValue - value ) / ( maxValue - minValue ) )
      else:
        red = minRed
        green = minGreen
        blue = minBlue
      newFillColor = QColor(red, green, blue)
      symbol.setFillColor(newFillColor)
      if renderer.drawPolygonOutline():
        # always black
        color = QColor(0,0,0)
        symbol.setColor(color)
      else:
        symbol.setColor(newFillColor)
      return symbol
     # # Continues Color: JUST TWO SYMBOLS FOR ALL CLASSES: CANNOT HANDLE THIS
     # # print renderer.symbols()
     # # value can be string or numbers...
     # if str(value)[0].isdigit():
     #   # oops it looks like a number  (TODO: use better test)
     #   # we should do a numeric comparison
     #   if float(value) == float(symbol.lowerValue()):
     #     return symbol
     # # else it's a string, we can do string compare
     # elif value == symbol.lowerValue():
     #   return symbol
    # else loop over symbols to find the current symbol for this feature
    value = (feature.attributeMap()[renderer.classificationAttributes()[0]]).toString()
    for symbol in renderer.symbols():
      lower = symbol.lowerValue()
      upper = symbol.upperValue()
      #print "lower: %s upper %s value = %s" % (lower, upper, value)
      if not lower and not upper:
          # default value for Unique Value
          default = symbol
      # only graduated classifications have lower AND uppervalues
      elif (float(value) >= float(symbol.lowerValue())) and (float(value) <= float(symbol.upperValue())):
        #print "TUSSEN TUSSEN:  %s tussen %s en %s " % (value, symbol.lowerValue(), symbol.upperValue())
        return symbol
    #print "RETURNING DEFAULT!!"
    return default

  def line2svg(self, feature, line, currentExtent, color, fillColor, pen):
    #print "calling line2svg..."
    linesvg = ['']
    svg = ''
    if self.svgType == SVG_TYPE_PATH:
      svg += '<path d="M '
    else:  # SVG_TYPE_SHAPE
      svg += '<polyline points="'
    lastPixel=[0,0]
    insideExtent = False
    coordCount = 0
    for point in line:
      if self.extentAsPoly.contains(point):
        insideExtent = True
      pixpoint =  self.w2p(point.x(), point.y(), self.iface.mapCanvas().mapUnitsPerPixel(), currentExtent.xMinimum(), currentExtent.yMaximum())
      #print pixpoint
      if lastPixel<>pixpoint:
        coordCount = coordCount +1
        if self.svgType==SVG_TYPE_PATH and coordCount>1:
          svg += 'L '
        svg += (str(pixpoint[0]) + ',' + str(pixpoint[1]) + ' ')
        lastPixel = pixpoint
    # if at least ONE pixel of this ring is in current view extent, return the area-string, otherwise return an empty string
    if not insideExtent:
      #print "RING FULLY OUTSIDE EXTENT: %s " % ring
      None
    else:
      # check if there are more then 2 coords: very small polygons on current map can have coordinates
      # which if rounded to pixels all come to the same pixel, resulting in just ONE x,y coordinate
      # we skip these
      if coordCount < 2:
        #print "Line contains just one pixel coordinate pair: skipping"
        None
      else:
        svg += '"/>\n'
        linesvg.append(svg)
    return linesvg

  # SHAPE-svg
  #  <polygon stroke="rgb(0,0,0)" fill="rgb(234,102,228)" stroke-width="1" points="439,238 445,230 ... 439,238"/>
  #     or
  #  <polygon points="439,238 445,230 ... 439,238"/>
  # PATH-svg
  #  <g stroke="rgb(0,0,0)" fill="rgb(234,102,228)" stroke-width="1" ><path d="M 439 238 L 445 220 ... L 439 238" /> </g>
  def polygon2svg(self, feature, polygon, currentExtent, color, fillColor, pen):
    #print "calling polygon2svg..."
    polygonsvg = ['']
    for ring in polygon:
      svg = ''
      if self.svgType == SVG_TYPE_PATH:
        svg += '<path d="M '
      else:  # SVG_TYPE_SHAPE
        svg = '<polygon points="'
      lastPixel=[0,0]
      insideExtent = False
      coordCount = 0
      for point in ring:
          if self.extentAsPoly.contains(point):
              insideExtent = True
          pixpoint =  self.w2p(point.x(), point.y(), self.iface.mapCanvas().mapUnitsPerPixel(), currentExtent.xMinimum(), currentExtent.yMaximum())
          if lastPixel<>pixpoint:
            coordCount = coordCount +1
            if self.svgType==SVG_TYPE_PATH and coordCount>1:
              svg += 'L '
            svg += (str(pixpoint[0]) + ',' + str(pixpoint[1]) + ' ')
            lastPixel = pixpoint
      # if at least ONE pixel of this ring is in current view extent, return the area-string, otherwise return an empty string
      if not insideExtent:
          #print "RING FULLY OUTSIDE EXTENT: %s " % ring
          None
      else:
          # check if there are more then 2 coords: very small polygons on current map can have coordinates
          # which if rounded to pixels all come to the same pixel, resulting in just ONE x,y coordinate
          # we skip these
          if coordCount < 2:
              #print "Ring contains just one pixel coordinate pair: skipping"
              None
          else:
              svg += '" />\n'
              polygonsvg.append(svg)
    return polygonsvg

  # NOT WORKING ????
  # pixpoint = m2p.transform(point.x(), point.y())
  # print m2p.transform(point.x(), point.y())
  # so for now: a custom 'world2pixel' method
  def w2p(self, x, y, mupp, minx, maxy):
    pixX = (x - minx)/mupp
    pixY = (y - maxy)/mupp
    return [int(pixX), int(-pixY)]

